#!/usr/bin/env python

from datetime import datetime
import logging

from bardolph.lib import injection
from bardolph.controller import i_controller


class Light:
    def __init__(self, name, group, location, color):
        self.name = name
        self.group = group
        self.location = location
        self.power = 12345
        self.color = color if color is not None else [0, 0, 0, 0]
        
    def __repr__(self):
        s = 'name: "{}", group: "{}", location: "{}", power: {}, '.format(
            self.name, self.group, self.location, self.power)
        s += 'color: {}'.format(self.color)
        return s
        
    def get_color(self):
        logging.info('Get color "{}" {}'.format(self.name, self.color))
        return self.color

    def set_color(self, color, duration = 0, _ = False):
        self.color = color
        logging.info('Color "{}" {}, {}'.format(self.name, color, duration))
        
    def set_return_color(self, color):
        self.color = color
        
    def set_power(self, power, duration):
        self.power = power
        logging.info('Power "{}", {}, {}'.format(self.name, power, duration))
        
    def get_power(self):
        return self.power
    
    def set_return_power(self, power):
        self.power = power
        
    def get_label(self):
        return self.name
    
    def get_location(self):
        return self.location
    
    def get_group(self):
        return self.group
  

class LightSet:
    def __init__(self):
        self.clear_lights()
        self.discover()
    
    def clear_lights(self):
        self.lights = {}
        self.groups = {}
        self.locations = {} 
        
    def discover(self):
        if len(self.lights) == 0:
            self.add_light('Table', 'Furniture', 'Home', [1, 2, 3, 4])
            self.add_light('Top',  'Pole', 'Home', [10, 20, 30, 40])
            self.add_light('Middle', 'Pole', 'Home', [100, 200, 300, 400])
            self.add_light(
                'Bottom', 'Pole', 'Home', [1000, 2000, 3000, 4000])
            self.add_light(
                'Chair', 'Furniture', 'Home', [10000, 20000, 30000, 40000])
     
    def add_light(self, name, group, location, color=None):
        new_light = Light(name, group, location, color)
        self.lights[name] = new_light
        if group is not None:
            if group in self.groups:
                self.groups[group].append(new_light)
            else:
                self.groups[group] = [new_light] 
        if location is not None:
            if location in self.locations:
                self.locations[location].append(new_light)
            else:
                self.locations[location] = [new_light]
    
    def get_light_names(self): 
        return self.lights.keys()
    
    def get_light(self, name):
        if not name in self.lights:
            logging.error("Light >\"{}\"< not in fake lights".format(name))
            return None
        return self.lights[name]
    
    def get_all_lights(self):
        return list(self.lights.values());

    def get_group_names(self):
        return self.groups.keys()
    
    def get_group(self, name):
        return self.groups.get(name, {})
    
    def get_location_names(self):
        return self.locations.keys()
    
    def get_location(self, name):
        return self.locations.get(name, {})

    def get_last_discover(self):
        return datetime.now()
    
    def get_successful_discovers(self):
        return 100
    
    def get_failed_discovers(self):
        return 10
    
    def set_color(self, color, duration):
        logging.info("Color (all) {}, {}".format(color, duration))

    def set_power(self, power_level, duration):
        logging.info("Power (all) {} {}".format(power_level, duration))


def configure():
    injection.bind_instance(LightSet()).to(i_controller.LightSet)



from enum import auto, Enum


class OpCode(Enum):
    COLOR = auto()
    END = auto()
    GET_COLOR = auto()
    NOP = auto()
    PAUSE = auto()
    POWER = auto()
    SET_REG = auto()
    STOP = auto()
    TIME_WAIT = auto()
    
    
class Operand(Enum):
    LIGHT = auto()
    GROUP = auto()
    LOCATION = auto()


class Instruction:
    def __init__(self, op_code = OpCode.NOP, name = None, param = None):
        self.op_code = op_code
        self.name = name
        self.param = param

    def __repr__(self):
        if self.param == None:
            if self.name == None:
                return 'Instruction(OpCode.{}, None, None)'.format(
                    self.op_code.name)
            else:
                return 'Instruction(OpCode.{}, "{}", None)'.format(
                    self.op_code.name, self.name)
        
        if type(self.param).__name__ == 'str':
            param_str = '"{}"'.format(self.param)
        else:
            param_str = str(self.param)
            
        return 'Instruction(OpCode.{}, "{}", {})'.format(
            self.op_code.name, self.name, param_str)
        
    def __eq__(self, other):
        if type(self) == type(other):
            return (self.op_code == other.op_code
                    and self.name == other.name and self.param == other.param)
        else:
            raise TypeError
        
    def as_list_text(self):
        if self.op_code != OpCode.SET_REG:
            return 'OpCode.{}'.format(self.op_code.name)
          
        if type(self.param).__name__ == 'str':
            param_str = '"{}"'.format(self.param)
        else:
            param_str = str(self.param)
            
        return 'OpCode.set_reg, "{}", {}'.format(self.name, param_str)



from enum import auto, Enum

class TokenTypes(Enum):
    ALL = auto()
    AND = auto()
    BRIGHTNESS = auto()
    DEFINE = auto()
    DURATION = auto()
    EOF = auto()
    GET = auto()
    GROUP = auto()
    HUE = auto()
    LITERAL = auto()
    LOCATION = auto()
    LOGICAL = auto()
    NUMBER = auto()
    OFF = auto()
    ON = auto()
    PAUSE = auto()
    KELVIN = auto()
    RAW = auto()
    SATURATION = auto()
    SET = auto()
    SYNTAX_ERROR = auto()
    TIME = auto()
    UNITS = auto()
    UNKNOWN = auto()



from . import i_lib
from . import injection


class Settings:
    active_config = None
    
    def __init__(self):
        self.refresh()
    
    def __str__(self):
        return str(self.configs)
    
    def __contains__(self, name):
        return self.active_config is not None and name in self.active_config
    
    def refresh(self):
        self.active_config = Settings.active_config
     
    def get_value(self, name, default = None):
        if self.active_config is None and default != None:
            return default
        elif default is None:
            return self.active_config[name]
        else:
            return self.active_config.get(name, default) 

    @classmethod
    def configure(cls):        
        Settings.active_config = None
        injection.bind(Settings).to(i_lib.Settings)
    
    @classmethod
    def specialize(cls, overrides):
        if Settings.active_config is None:
            Settings.active_config = overrides.copy()
        else:
            Settings.active_config.update(overrides)

    @classmethod
    def put_config(cls, config):
        Settings.active_config = config


class Base:
    def __init__(self, overrides):
        self.overrides = overrides

    def and_override(self, override):
        return Overrider(self, override)
        
    def configure(self):
        Settings.configure()
        Settings.specialize(self.overrides)
        

class Overrider:
    def __init__(self, base, override):
        self.base = base
        self.override = override
        
    def configure(self):
        self.base.configure()
        Settings.specialize(self.override)


def using_base(overrides):
    return Base(overrides)


def configure():
    Settings.configure()


def specialize(overrides):
    return Settings.specialize(overrides)



import threading
import time

from . import injection
from . import i_lib


def now():
    # seconds
    return time.time()


def as_seconds(t):
    return t  / 1000.0


def as_ms(t):
    return t * 1000.0

    
def configure():
    injection.bind(Clock).to(i_lib.Clock)


# All time quantities are in seconds.
class Clock(i_lib.Clock):
    def __init__(self):
        self.keep_going = False
        self.event = threading.Event()
        self.start_time = 0.0
        self.cue_time = 0.0
        self.keep_going = True
    
    def start(self):
        self.reset()
        threading.Thread(target=self.run, args=(), daemon=True).start()
    
    @injection.inject(i_lib.Settings)
    def run(self, settings):
        self.keep_going = True
        sleep_time = float(settings.get_value('sleep_time'))
        while self.keep_going:
            if sleep_time > 0.0:
                time.sleep(sleep_time)
            self.fire()
    
    def stop(self):
        self.keep_going = False
    
    def reset(self):
        self.cue_time = 0.0
        self.start_time = now()
        
    def et(self):
        return (time.time() - self.start_time)

    def fire(self):
        if self.keep_going:
            self.event.set()
            self.event.clear()
        
    def wait(self):
        if self.keep_going:
            self.event.wait()
        return self.keep_going

    def pause_for(self, delay):
        self.cue_time += delay
        while self.et() < self.cue_time:
            self.wait()
        



#!/usr/bin/env python

import argparse

from ..lib import injection
from ..lib import settings
from ..parser.token_types import TokenTypes

from . import config_values
from . import light_module
from .i_controller import LightSet
from .lsc import Compiler
from .units import Units


def quote_if_string(param):
    return '{}'.format(param) if isinstance(param, str) else param

class Snapshot:
    def start_snapshot(self): pass
    def start_light(self, light): pass
    def handle_setting(self, name, value): pass
    def handle_power(self, power): pass
    def end_light(self): pass
    def end_snapshot(self): pass
    
    def handle_color(self, color):
        self.handle_setting('hue', color[0])
        self.handle_setting('saturation', color[1])
        self.handle_setting('brightness', color[2])
        self.handle_setting('kelvin', color[3])
    
    @injection.inject(LightSet)
    def generate(self, light_set):
        self.start_snapshot()
        light_names = light_set.get_light_names()
        for name in light_names:
            light = light_set.get_light(name)
            self.start_light(light)
            self.handle_color(light.get_color())
            self.handle_power(light.get_power())
            self.end_light()
        self.end_snapshot()
        return self        


class ScriptSnapshot(Snapshot):
    """ Generate a .ls script. Ignore power on/off. """
    def __init__(self):
        self.light_name = ''
        self.power = True
        self.script = ''
    
    def start_snapshot(self):
        self.script = 'raw duration 1500\n'
        
    def start_light(self, light):
        self.light_name = light.get_label()
        
    def handle_setting(self, name, value):
        self.script += '{} {} '.format(name, value)
        
    def handle_color(self, color):
        params = zip([
            TokenTypes.HUE, TokenTypes.SATURATION, TokenTypes.BRIGHTNESS,
            TokenTypes.KELVIN
            ], color)
        u = Units()
        for param in params:
            reg, value = param
            fmt = '{} {:.2f} ' if u.requires_conversion(reg) else '{} {} '
            self.script += fmt.format(
                reg.name.lower(), u.as_logical(reg, value))
                        
    def handle_power(self, power):
        self.power = power
          
    def end_light(self):
        self.script += 'set "{}"\n'.format(self.light_name)     
        fmt = 'on "{}"\n' if self.power else 'off "{}"\n'
        self.script += fmt.format(self.light_name)

    def get_text(self):
        return '{}\n'.format(self.script)
 
     
class InstructionSnapshot(Snapshot):
    """ Generate a list of lists, one for each light. """
    def __init__(self):
        self.light_name = ''
        self.snapshot = ''
        self.power = ''
    
    def start_snapshot(self):
        self.snapshot = ''
        
    def start_light(self, light):
        self.light_name = light.get_label()
    
    def handle_setting(self, name, value):
        self.snapshot += 'OpCode.set_reg, "{}", {},\n'.format(
            name, quote_if_string(value))
    
    def handle_power(self, power):
        self.power = power
        
    def end_light(self):
        self.snapshot += 'OpCode.set_reg, "name", "{}",\n'.format(
            self.light_name)
        self.snapshot += 'OpCode.set_reg, "operand", Operand.light,\n'        
        self.snapshot += 'OpCode.color,\n'
        self.snapshot += 'OpCode.set_reg, "power", {},\n'.format(self.power)
        self.snapshot += 'OpCode.power,\n'.format(self.power)
        
    def get_text(self):
        return self.snapshot[:-2]
    
    
class TextSnapshot(Snapshot):
    """ Generate plain text. """
    def __init__(self):
        self.field_width = len('saturation  ')
        self.text = ''
        self.add_field('name').add_field('hue')
        self.add_field('saturation').add_field('brightness')
        self.add_field('kelvin').add_field('power')
        self.text += '\n'
        self.text += '-' * ((self.field_width - 1) * 6)
        self.text += '\n'
        
    def generate(self):
        super().generate()
        self.add_sets()
        return self
    
    def add_field(self, data):
        self.text += str(data).ljust(self.field_width)
        return self
                  
    def start_light(self, light):
        self.add_field(light.get_label())
        
    def handle_color(self, color):
        params = zip([
            TokenTypes.HUE, TokenTypes.SATURATION, TokenTypes.BRIGHTNESS,
            TokenTypes.KELVIN
            ], color)
        u = Units()
        for p in params:
            self.add_field('{:>6.2f}'.format(u.as_logical(p[0], p[1])))
            
    def handle_power(self, power):
        self.add_field('{:>5d}'.format(power))
            
    def end_light(self):
        self.text += '\n'    
        
    def get_text(self):
        return self.text
    
    @injection.inject(LightSet)
    def add_sets(self, lights):
        self.add_set('Groups', lights.get_group_names, lights.get_group)
        self.add_set(
            'Locations', lights.get_location_names, lights.get_location)
        
    def add_set(self, heading, name_fn, get_fn):
        self.text += '\n{}\n'.format(heading)
        self.text += '-' * 17
        self.text += '\n'
        for name in name_fn():
            self.text += '{}\n'.format(name)
            for light in get_fn(name):
                self.text += '   {}\n'.format(light.get_label())
                

class DictSnapshot(Snapshot):
    """ Generate a list of dictionaries, one for each light. """
    def __init__(self):
        self.snapshot = None
        self.current_dict = None
    
    def start_snapshot(self):
        self.snapshot = []
        self.current_dict = {}        
    
    def start_light(self, light):
        self.current_dict = {'name': light.get_label()}
          
    def handle_setting(self, name, value):
        self.current_dict[name] = value
    
    def end_light(self):
        self.snapshot.append(self.current_dict)
    
    def get_text(self):
        return str(self.get_list())
    
    def get_list(self):
        return self.snapshot


def do_gen(ctor):
    print(ctor().generate().get_text() + '\n')
        
        
def main():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '-l', '--list', help='output instruction list', action='store_true')
    ap.add_argument(
        '-d', '--dict', help='output dictionary format', action='store_true')
    ap.add_argument(
        '-p', '--py', help='output Python file', action='store_true')
    ap.add_argument(
        '-s', '--script', help='output script format', action='store_true')
    ap.add_argument(
        '-t', '--text', help='output text format', action='store_true')
    args = ap.parse_args()
    do_script = args.script
    do_dict = args.dict
    do_list = args.list
    do_py = args.py
    do_text = args.text or (not (do_py or do_script or do_dict or do_list))  

    injection.configure()   
    settings.using_base(config_values.functional).and_override(
        {'single_light_discover': True}).configure()
    light_module.configure()
    
    if do_dict:
        do_gen(DictSnapshot)
    if do_list:
        do_gen(InstructionSnapshot)
    if do_script:
        do_gen(ScriptSnapshot)
    if do_text:
        do_gen(TextSnapshot)
    if do_py:
        text = InstructionSnapshot().generate().get_text()
        Compiler().generate_from(text)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import argparse
import logging

from ..lib import injection
from ..lib import settings

from . import config_values
from .instruction import Instruction, OpCode, Operand 
from . import light_module
from . import machine

instructions = [
    #instructions
]

current_instruction = 0

def next_instruction():
    global current_instruction
    if current_instruction < len(instructions):
        value = instructions[current_instruction]    
        current_instruction += 1
        return value
    else: 
        return None

def build_instructions():
    program = []
    op_code = next_instruction()
    while op_code != None:
        if op_code == OpCode.set_reg:
            name = next_instruction()
            value = next_instruction()
            program.append(Instruction(OpCode.set_reg, name, value))
        else:
            program.append(Instruction(op_code))
        op_code = next_instruction()
    return program

def run_script(overrides = None):
    injection.configure()   
    settings.using_base(config_values.functional).configure()
    if overrides != None:
        settings.specialize(overrides)

    light_module.configure()    
    machine.Machine().run(build_instructions())

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '-d', '--debug', help='do debug-level logging', action='store_true')
    ap.add_argument(
        '-f', '--fakes', help='use fake lights', action='store_true')
    args = ap.parse_args()
    
    overrides = { 'sleep_time': 0.1 }
    if args.debug:
        overrides['log_level'] = logging.DEBUG
        overrides['log_to_console'] = True
    if args.fakes:
        overrides['use_fakes'] = True
    run_script(overrides)


if __name__ == '__main__':
    main()



from enum import auto, Enum

from ..parser.token_types import TokenTypes


HSB_RANGE = (0, 65535)
KELVIN_RANGE = (2500, 9000)
RAW_RANGE = (0, 65535)


class UnitMode(Enum):
    LOGICAL = auto()
    RAW = auto()


class Units:
    def has_range(self, reg):
        return reg not in (TokenTypes.DURATION, TokenTypes.TIME)
    
    def requires_conversion(self, reg):
        return reg not in (
            TokenTypes.DURATION, TokenTypes.KELVIN, TokenTypes.TIME)

    def get_range(self, reg):
        """
        Return the allowable range, in raw units, for a parameter.
        
        Reg:
            token_type.TokenType designating the register to be set.
        
        Returns:
            A tuple containing (minimum, maximum), or None if the parameter
            does not have a limited range of values.
        """
        return {
            TokenTypes.DURATION: None,
            TokenTypes.TIME: None,
            TokenTypes.KELVIN: KELVIN_RANGE }.get(reg, RAW_RANGE)
        
    def as_raw(self, reg, logical_value):
        """If necessary, converts to integer value that can be passed into the
        light API.
        
        Args:
            reg: TokenType corresponding to the register being set.
            logical_value: the number to be converted.
            
        Returns:
            If no conversion is done, returns the incoming value untouched.
            Otherwise, an integer that corresponds to the logical value.
        """
        value = logical_value
        if self.requires_conversion(reg):
            if reg == TokenTypes.HUE:
                if logical_value == 360.0 or logical_value == 0.0:
                    value = 0;
                else:
                    value = round((logical_value % 360.0) / 360.0 * 65535.0)
            elif reg in (TokenTypes.BRIGHTNESS, TokenTypes.SATURATION):
                if logical_value == 100.0:
                    value = 65535
                else:
                    value = round(logical_value / 100.0 * 65535.0)            
        return round(value)

    def as_logical(self, reg, raw_value):
        """If necessary, converts to floating-point logical value that 
        typically apears in a script.
        
        Args:
            reg: TokenType corresponding to the register being set.
            raw_value: the number to be converted.
            
        Returns:
            If no conversion is done, returns the incoming value untouched.
            Otherwise, a float that corresponds to the raw value.
        """
        value = raw_value
        if not self.requires_conversion(reg):
            return raw_value
        if reg == TokenTypes.HUE:
            value = float(raw_value) / 65535.0 * 360.0
        elif reg in (TokenTypes.BRIGHTNESS, TokenTypes.SATURATION):
            if raw_value == 65535:
                value = 100.0
            else:
                value = float(raw_value) / 65535.0 * 100.0
        return value



import logging

functional = {
    'default_num_lights': 5,
    'sleep_time': 0.01, # seconds
    
    'generated_path': 'generated',
    'log_date_format': '%D %H:%M:%S',
    'log_format':
        '%(asctime)s %(filename)s(%(lineno)d) %(funcName)s(): %(message)s',
    'log_level': logging.WARNING,
    'log_to_console': True,
    'refresh_sleep_time': 600, # seconds
    'failure_sleep_time': 120, # seconds
    'single_light_discover': False,
    'use_fakes': False,

    # Ignored unless log_to_console gets overridden to False.
    'log_file_name': '/var/log/lights/lights.log'
}




#!/usr/bin/env python

import argparse
import logging

from ..lib import injection
from ..lib import job_control
from ..lib import settings

from . import light_module
from . import config_values
from .script_runner import ScriptRunner

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument('file', help='name of the script file', nargs='*')
    ap.add_argument(
        '-f', '--fakes', help='use fake lights', action='store_true')
    ap.add_argument(
        '-r', '--repeat', help='repeat until key pressed', action='store_true')
    ap.add_argument(
        '-s', '--script', help='run script from command line', action='store')
    ap.add_argument(
        '-v', '--verbose', help='verbose output', action='store_true')
    args = ap.parse_args()
    
    injection.configure()   
    settings.using_base(config_values.functional).configure()
    overrides = {
        'log_date_format': "%I:%M:%S %p",
        'log_format': '%(asctime)s %(filename)s(%(lineno)d): %(message)s',
        'log_level': logging.DEBUG if args.verbose else logging.INFO,
        'log_to_console': True,
        'sleep_time': 0.1
    }
    if args.fakes:
        overrides['use_fakes'] = True
        
    settings.specialize(overrides)
    light_module.configure()

    jobs = job_control.JobControl(args.repeat)
    if args.script is not None:
        jobs.add_job(ScriptRunner.from_string(args.script))   
    for file_name in args.file:
        jobs.add_job(ScriptRunner.from_file(file_name))


if __name__=="__main__":
    main()



from bardolph.lib import injection, settings
from bardolph.controller import config_values, light_module
from bardolph.lib import job_control
from bardolph.controller.script_runner import ScriptRunner

class LsModule:
    jobs = None
    
    def configure(self):
        injection.configure()   
        settings.using_base(config_values.functional).configure()
        light_module.configure()
        LsModule.jobs = job_control.JobControl()

    def queue_script(self, script):
        LsModule.jobs.add_job(ScriptRunner.from_string(script))
    
    def request_stop(self):
        LsModule.jobs.request_stop()

def configure():
    LsModule().configure()
    
def queue_script(script):
    LsModule().queue_script(script)
    
def request_stop():
    LsModule().request_stop()



import logging

from ..lib.i_lib import Clock
from ..lib.color import average_color
from ..lib.injection import inject, provide

from .get_key import getch
from .i_controller import LightSet
from .instruction import OpCode, Operand


class Registers:
    def __init__(self):
        self.hue = 0
        self.saturation = 0
        self.brightness = 0
        self.kelvin = 0
        self.duration = 0
        self.power = False
        self.name =  None
        self.operand = None
        self.time = 0 # ms.

    def get_color(self):
        return [self.hue, self.saturation, self.brightness, self.kelvin]

        
class Machine:
    def __init__(self):
        self.pc = 0
        self.cue_time = 0
        self.clock = provide(Clock)
        self.variables = {}
        self.program = []
        self.reg = Registers()
        self.enable_pause = True
        self.fn_table = {
            OpCode.COLOR: self.color,
            OpCode.END: self.end,
            OpCode.GET_COLOR: self.get_color,
            OpCode.NOP: self.nop,
            OpCode.PAUSE: self.pause,
            OpCode.POWER: self.power,
            OpCode.SET_REG: self.set_reg,
            OpCode.STOP: self.stop,
            OpCode.TIME_WAIT: self.time_wait
        }

    def run(self, program):
        self.program = program
        self.pc = 0
        self.cue_time = 0  
        self.clock.start()
        while self.pc < len(self.program):
            inst = self.program[self.pc]
            if inst.op_code == OpCode.STOP:
                break
            self.fn_table[inst.op_code]()
            self.pc += 1
        self.clock.stop()

    def apply_to(self, get_fn, op_fn):
        lst_or_obj = get_fn(self.reg.name)
        if lst_or_obj is not None:
            try:
                for light in lst_or_obj:
                    op_fn(light)
            except TypeError:
                op_fn(lst_or_obj)

    def color(self): {
            Operand.LIGHT: self.color_light, 
            Operand.GROUP: self.color_group,
            Operand.LOCATION: self.color_location
        }[self.reg.operand]()

    @inject(LightSet)
    def color_light(self, light_set):
        if self.reg.name is None:
            light_set.set_color(self.reg.get_color(), self.reg.duration)
        else:
            self.apply_to(self.one_light, self.light_command)
    
    @inject(LightSet)
    def color_group(self, light_set):
        self.apply_to(light_set.get_group, self.light_command)
        
    @inject(LightSet)
    def color_location(self, light_set):
        self.apply_to(light_set.get_location, self.light_command)
        
    def light_command(self, target):
        """ target is lifxlan.light.Light """
        target.set_color(self.reg.get_color(), self.reg.duration, True)
                                                   
    def light_set_command(self, target):
        """ target is lib.i_lib.LightSet. """
        target.set_color(self.reg.get_color(), self.reg.duration)
                                                   
    @inject(LightSet)
    def one_light(self, name, light_set):
        light = light_set.get_light(name)
        if light is None:
            self.report_missing(name)
            return None
        else:
            return light
        
    def power(self): { 
        Operand.LIGHT: self.power_light, 
        Operand.GROUP: self.power_group,
        Operand.LOCATION: self.power_location
    }[self.reg.operand]()
    
    @inject(LightSet)
    def power_light(self, light_set):
        if self.reg.name is None:
            light_set.set_power(self.power_param(), self.reg.duration)
        else:
            self.apply_to(self.one_light, self.power_command)
                
    @inject(LightSet)
    def power_group(self, light_set):
        self.apply_to(light_set.get_group, self.power_command)
        
    @inject(LightSet)
    def power_location(self, light_set):
        self.apply_to(light_set.get_location, self.power_command)           

    def power_command(self, target):
        """ target can be lifxlan.light.Light or lib.i_lib.LightSet. """
        target.set_power(self.power_param(), self.reg.duration)

    def get_color(self): { 
        Operand.LIGHT: self.get_light_color, 
        Operand.GROUP: self.get_group_color,
        Operand.LOCATION: self.get_location_color
    }[self.reg.operand]()
    
    @inject(LightSet)
    def get_light_color(self, light_set):
        if self.reg.name is None:
            self.get_average_color(light_set.get_all_lights)
        else:
            light = light_set.get_light(self.reg.name)
            if light is None:
                self.report_missing(self.reg.name)
            else:
                self.color_to_reg(light.get_color())

    @inject(LightSet)
    def get_group_color(self, light_set):
        self.get_average_color(light_set.get_group)
        
    @inject(LightSet)
    def get_location_color(self, light_set):
        self.get_average_color(light_set.get_location)
    
    def get_average_color(self, get_fn):
        if self.reg.name is None:
            colors = [light.get_color() for light in get_fn()]
        else:
            colors = get_fn(self.reg.name)
        self.color_to_reg(average_color(colors))
            
    def nop(self): pass
        
    def pause(self):
        if self.enable_pause:
            print("Press any to continue, q to quit, ! to run.")
            c = getch()
            if c == 'q':
                self.stop()
            else:
                print("Running...")
                if c == '!':
                    self.enable_pause = False
            
    def check_wait(self):
        time = self.reg.time
        if time > 0:
            self.clock.pause_for(time / 1000.0)

    def end(self):
        self.stop()

    def stop(self):
        self.pc = len(self.program)
    
    def set_reg(self):
        inst = self.program[self.pc]
        setattr(self.reg, inst.name, inst.param)
        
    def time_wait(self):
        self.check_wait()
          
    def report_missing(self, name):
        logging.warning("Light \"{}\" not found.".format(name))

    def power_param(self):
        return 65535 if self.reg.power else 0
        
    def color_from_reg(self):
        return [self.reg.hue, self.reg.saturation, self.reg.brightness,
            self.reg.kelvin]

    def color_to_reg(self, color):
        if color != None:
            r = self.reg
            r.hue, r.saturation, r.brightness, r.kelvin = color



import sys
import termios
import tty

def getch():
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(fd)
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch


#!/usr/bin/env python

import logging
import threading
import time

import lifxlan

from ..lib.color import average_color
from ..lib.injection import bind_instance, inject
from ..lib.i_lib import Settings

from . import i_controller 


class LightSet(i_controller.LightSet):
    the_instance = None

    def __init__(self):
        self.light_dict = {}
        self.groups = {}
        self.locations = {}
        self.num_successful_discovers = 0
        self.num_failed_discovers = 0
        self.last_discover = None
   
    def __len__(self):
        return self.get_count()

    @classmethod
    def configure(cls):
        if LightSet.the_instance == None:
            LightSet.the_instance = LightSet()

    @classmethod
    def get_instance(cls):
        return LightSet.the_instance
    
    @inject(Settings)
    def discover(self, settings):
        expected = settings.get_value('default_num_lights', 0)       
        new_dict = {}
        try:
            if expected == 0:
                light_list = self.broadcast_state_service()
            else:
                light_list = self.broadcast_state_service(expected)
            for light in light_list:
                new_dict[light.get_label()] = light  
            self.light_dict = new_dict
            self.last_discover = time.localtime()
            
            self.groups = self.build_set(
                light_list, lifxlan.device.Device.get_group)
            self.locations = self.build_set(
                light_list, lifxlan.device.Device.get_location)
        except lifxlan.errors.WorkflowException as ex:
            logging.error("Error during discovery {}".format(ex))
            return False
              
        actual = len(self.light_dict) 
        if actual < expected:
            logging.warning(
                "Expected {} devices, found {}".format(expected, actual))
            return False
        else:
            return True
        
    def broadcast_state_service(self, num_expected):
        lifx = lifxlan.LifxLAN(num_expected)
        responses = lifx.broadcast_with_resp(
            lifxlan.msgtypes.GetService, lifxlan.msgtypes.StateService,
            max_attempts = 5)

        light_list = list()
        for response in responses:
            new_light = lifxlan.light.Light(
                response.target_addr, response.ip_addr, response.service,
                response.port, lifx.source_id, lifx.verbose)
            light_list.append(new_light)
        
        return light_list
       
    def build_set(self, light_list, fn):
        # Produces a dictionary keyed on group or location name, pointing to
        # a list of (lifxlan) light objects.
        sets = {}
        for light in light_list:
            set_name = fn(light)
            if set_name in sets:
                sets[set_name].append(light)
            else:
                sets[set_name] = [light]
        return sets

    def get_light_names(self):
        """ list of strings """
        return self.light_dict.keys()
        
    def get_light(self, name):
        """ returns an instance of lifxlan.Light, or None if it's not there """ 
        return self.light_dict.get(name, None)
    
    def get_group_names(self):
        """ list of strings """
        return self.groups.keys()
        
    def get_group(self, name):
        """ list of Lights """ 
        return self.groups.get(name, [])
    
    def get_location_names(self):
        """ list of strings """
        return self.locations.keys()
        
    def get_location(self, name):
        """ list of Lights. """ 
        return self.locations.get(name, [])
    
    def get_all_lights(self):
        """ list of Lights. """
        return self.light_dict.values()
    
    def get_count(self):
        return len(self.light_dict)
    
    def get_last_discover(self):
        return self.last_discover
    
    def get_successful_discovers(self):
        return self.num_successful_discovers
    
    def get_failed_discovers(self):
        return self.num_failed_discovers
    
    @inject(Settings)
    def set_color(self, color, duration, settings):
        num_expected = settings.get_value('default_num_lights')
        lifx = lifxlan.LifxLAN(num_expected)
        lifx.set_color_all_lights(color, duration, True)
        return True
    
    def get_color(self):
        """ Returns arithmetic mean of each color setting. """
        return average_color(
            list(self.get_lifxlan().get_color_all_lights().values()))
    
    def set_power(self, power_level, duration):
        self.get_lifxlan().set_power_all_lights(power_level, duration, True)
        return True
    
    def get_power(self):
        """ Returns True if at least one bulb is on. """
        return self.any_power(
            self.get_lifxlan().get_power_all_lights())
            
    def any_power(self, state_dict):
        for state in state_dict.values():
            if state:
                return True
        return False
    
    @inject(Settings)
    def get_lifxlan(self, settings):
        num_expected = settings.get_value('default_num_lights')
        return lifxlan.LifxLAN(num_expected)

def start_light_refresh():
    logging.info("Starting refresh thread.")
    threading.Thread(
        target=light_refresh, name='rediscover', daemon=True).start()


@inject(Settings)
def light_refresh(settings):
    logging.debug("will start discover loop")
    
    success_sleep_time = settings.get_value('refresh_sleep_time')
    failure_sleep_time = settings.get_value('failure_sleep_time')
    complete_success = False
    
    while True:
        logging.info("Discovering light_set.")
        lights = LightSet.get_instance()
        try:
            complete_success = lights.discover()
            lights.num_successful_discovers += 1
        except lifxlan.errors.WorkflowException as ex:
            logging.warning("Error during discovery {}".format(ex))
            lights.num_failed_discovers += 1
        
        time.sleep(
            success_sleep_time if complete_success else failure_sleep_time)
        
    logging.info("Exiting light_refresh()")
    

@inject(Settings)
def configure(settings):
    LightSet.configure()
    lights = LightSet.get_instance()
    bind_instance(lights).to(i_controller.LightSet)
    lights.discover()
    
    single = settings.get_value('single_light_discover')
    logging.debug("single_light_discover: {}".format(single))
    if (not single):
        start_light_refresh()



#!/usr/bin/env python

import argparse
import os
import stat

from . import light_module
from ..lib import injection
from ..lib import settings

from . import config_values
from .script_runner import ScriptRunner

class Compiler:      
    def generate(self, file_name):
        program = ScriptRunner.from_file(file_name).program
        text = ",\n".join(map(lambda inst: inst.as_list_text(), program))
        self.generate_from(text, file_name)
                
    def generate_from(self, instruction_text, file_name=None):
        output = ""
        with open(
            os.path.join('bardolph', 'controller', 'lsc_template.py')) as f:
            for line in f:
                if line.find("#instructions") > -1:
                    output += instruction_text
                else:
                    output += line
        if file_name is not None:
            output += "\n# Generated by lsc.py from {}\n".format(file_name)

        output_name = '__generated__.py'
        out_file = open(output_name, 'w')
        out_file.write(output)
        out_file.close()
        os.chmod(output_name, stat.S_IRWXU | stat.S_IRGRP | stat.S_IROTH)

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("file", help="name of the script file", nargs='+')
    args = ap.parse_args()
    
    injection.configure()   
    settings.using_base(config_values.functional).configure()
    light_module.configure()
    
    compiler = Compiler()
    for file_name in args.file:
        print(file_name)
        compiler.generate(file_name)

if __name__=="__main__":
    main()




#!/usr/bin/env python

import argparse
import logging
import re

from ..controller.instruction import Instruction, OpCode, Operand
from ..controller.units import UnitMode, Units
from . import lex
from .token_types import TokenTypes


word_regex = re.compile(r"\w+")


class Parser:
    def __init__(self):
        self.lexer = None
        self.error_output = ''
        self.light_state = {}
        self.name = None
        self.current_token_type = None
        self.current_token = None
        self.op_code = OpCode.NOP
        self.symbol_table = {}
        self.code = []
        self.unit_mode = UnitMode.LOGICAL
        
    def parse(self, input_string):
        self.code.clear()
        self.error_output = ''
        self.lexer = lex.Lex(input_string)
        self.next_token()
        success = self.script()
        self.lexer = None
        if not success:
            return None
        self.optimize()
        return self.code

    def load(self, file_name):
        logging.debug('File name: {}'.format(file_name))
        try:
            f = open(file_name, 'r')
            input_string = f.read()
            f.close()
            return self.parse(input_string)
        except FileNotFoundError:
            logging.error('Error: file {} not found.'.format(file_name))
        except:
            logging.error('Error accessing file {}'.format(file_name))

    def script(self):
        return self.body() and self.eof()
        
    def body(self):
        succeeded = True
        while succeeded and self.current_token_type != TokenTypes.EOF:
            succeeded = self.command()
        return succeeded

    def eof(self):
        if self.current_token_type != TokenTypes.EOF:
            return self.trigger_error("Didn't get to end of file.")
        return True

    def command(self):
        fn = {
            TokenTypes.BRIGHTNESS: self.set_reg,
            TokenTypes.DEFINE: self.definition,
            TokenTypes.DURATION: self.set_reg,
            TokenTypes.GET: self.get,
            TokenTypes.HUE: self.set_reg,
            TokenTypes.KELVIN: self.set_reg,
            TokenTypes.OFF: self.power_off,
            TokenTypes.ON: self.power_on,
            TokenTypes.PAUSE: self.pause,
            TokenTypes.SATURATION: self.set_reg,
            TokenTypes.SET: self.set,
            TokenTypes.TIME: self.set_reg,
            TokenTypes.UNITS: self.set_units,
        }.get(self.current_token_type, self.syntax_error)
        return fn() if fn != None else True
        
    def set_reg(self):
        self.name = self.current_token
        reg = self.current_token_type
        self.next_token()

        if self.current_token_type == TokenTypes.NUMBER:
            try:
                value = round(float(self.current_token))
            except ValueError:
                return self.token_error('Invalid number: "{}"')
        elif self.current_token_type == TokenTypes.LITERAL:
            value = self.current_token
        elif self.current_token in self.symbol_table:
            value = self.symbol_table[self.current_token]
        else:
            return self.token_error('Unknown parameter value: "{}"')

        u = Units()
        if self.unit_mode == UnitMode.LOGICAL:
            value = u.as_raw(reg, value)
            if u.has_range(reg):
                (min_val, max_val) = u.get_range(reg)
                if value < min_val or value > max_val:
                    if self.unit_mode == UnitMode.LOGICAL:
                        min_val = u.as_logical(reg, min_val)
                        max_val = u.as_logical(reg, max_val)
                    return self.trigger_error(
                        '{} must be between {} and {}'.format(
                            reg.name.lower(), min_val, max_val))

        self.add_instruction(OpCode.SET_REG, self.name, value)
        return self.next_token()
    
    def set_units(self):
        self.next_token()
        mode = {
            TokenTypes.RAW: UnitMode.RAW,
            TokenTypes.LOGICAL:UnitMode.LOGICAL
        }.get(self.current_token_type, None)
    
        if mode is None:
            return self.trigger_error(
                'Invalid parameter "{}" for units.'.format(self.current_token))

        self.unit_mode = mode
        return self.next_token()

    def set(self):
        return self.action(OpCode.COLOR)
    
    def get(self):
        return self.action(OpCode.GET_COLOR)
    
    def power_on(self):
        self.add_instruction(OpCode.SET_REG, 'power', True)
        return self.action(OpCode.POWER)
        
    def power_off(self):
        self.add_instruction(OpCode.SET_REG, 'power', False)
        return self.action(OpCode.POWER)
    
    def pause(self):
        self.add_instruction(OpCode.PAUSE)
        self.next_token()
        return True
        
    def action(self, op_code):
        self.op_code = op_code
        self.next_token()

        if self.current_token_type == TokenTypes.GROUP:
            self.add_instruction(OpCode.SET_REG, 'operand', Operand.GROUP)
            self.next_token()
        elif self.current_token_type == TokenTypes.LOCATION:
            self.add_instruction(OpCode.SET_REG, 'operand', Operand.LOCATION)
            self.next_token()
        else:
            self.add_instruction(OpCode.SET_REG, 'operand', Operand.LIGHT)
        
        return self.operand_list()
    
    def operand_list(self):
        if self.current_token_type == TokenTypes.ALL:
            self.add_instruction(OpCode.SET_REG, 'name', None)
            if self.op_code != OpCode.GET_COLOR:
                self.add_instruction(OpCode.TIME_WAIT)
            self.add_instruction(self.op_code) 
            return self.next_token()
        
        if not self.operand_name():
            return False
        
        self.add_instruction(OpCode.SET_REG, 'name', self.name)
        if self.op_code != OpCode.GET_COLOR:
            self.add_instruction(OpCode.TIME_WAIT)
        self.add_instruction(self.op_code)
        while self.current_token_type == TokenTypes.AND:
            if not self.and_operand():
                return False
        return True
    
    def operand_name(self):
        if self.current_token_type == TokenTypes.LITERAL:
            self.name = self.current_token
        elif self.current_token in self.symbol_table:
            self.name = self.symbol_table[self.current_token]
        else:
            return self.token_error('Unknown variable: {}')
        return self.next_token()

    def and_operand(self):
        self.next_token()
        if not self.operand_name():
            return False
        self.add_instruction(OpCode.SET_REG, 'name', self.name)
        self.add_instruction(self.op_code)
        return True
           
    def definition(self):
        self.next_token()
        if self.current_token_type in [
                TokenTypes.LITERAL, TokenTypes.NUMBER]:
            return self.token_error('Unexpected literal: {}')
        
        var_name = self.current_token
        self.next_token()
        if self.current_token_type == TokenTypes.NUMBER:
            value = int(self.current_token)
        elif self.current_token_type == TokenTypes.LITERAL:
            value = self.current_token
        elif self.current_token in self.symbol_table:
            value = self.symbol_table[self.current_token]
        else:
            return self.token_error('Unknown term: "{}"')

        self.symbol_table[var_name] = value
        self.next_token()
        return True

    def add_instruction(self, op_code, name = None, param = None):
        self.code.append(Instruction(op_code, name, param))

    def get_errors(self):
        return self.error_output;
    
    def add_message(self, message):
        self.error_output += '{}\n'.format(message)
        
    def trigger_error(self, message):
        full_message = 'Line {}: {}'.format(self.lexer.line_num, message)
        logging.error(full_message)
        self.add_message(full_message)
        return False
    
    def token_error(self, message_format):
        return self.trigger_error(message_format.format(self.current_token))

    def next_token(self):
        (self.current_token_type, self.current_token) = self.lexer.next_token()
        return True
        
    def syntax_error(self):
        return self.token_error('Unexpected input "{}"')

    def optimize(self):
        """
        Eliminate an instruction if it would set a register to the same value
        that was assigned to it in the previous SET_REG instruction.
        
        Any GET_COLOR instruction clears out the previous value cache. 
        """
        opt = []
        prev_value = {}
        for inst in self.code:
            if inst.op_code == OpCode.GET_COLOR:
                prev_value = {}
            if inst.op_code != OpCode.SET_REG:
                opt.append(inst)
            else:
                if inst.name not in prev_value or (
                        prev_value[inst.name] != inst.param):
                    opt.append(inst)
                    prev_value[inst.name] = inst.param
        self.code = opt
                
                
def main():
    ap = argparse.ArgumentParser()
    ap.add_argument('file', help='name of the script file')
    args = ap.parse_args()
    
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(filename)s(%(lineno)d) %(funcName)s(): %(message)s')
    parser = Parser()
    output_code = parser.load(args.file)
    if output_code:
        for inst in output_code:
            print(inst)
    else:
        print(parser.error_output)

    
if __name__ == '__main__':
    main()
    
"""
    <script> ::= <body> <EOF>
    <body> ::= <command> *
    <command> ::=
        "brightness" <set_reg>
        | "define" <definition>
        | "duration" <set_reg> 
        | "hue" <set_reg> 
        | "kelvin" <set_reg>
        | "off" <power_off>
        | "on" <power_on>
        | "pause" <pause>
        | "saturation" <set_reg>
        | "set" <set>
        | "units" <set_units>
        | "time" <set_reg>
    <set_reg> ::= <name> <number> | <name> <literal> | <name> <symbol>
    <set> ::= <action>
    <get> ::= <action>
    <power_off> ::= <action>
    <power_on> ::= <action>
    <action> ::= <op_code> <operand_list>
    <operand_list> ::= "all" | <operand_name> | <operand_name> <and_operand> *
    <operand_name> ::= <token>
    <and_operand> ::= "and" <operand_name>
    <definition> ::= <token> <number> | <token> <literal>
    <literal> ::= "\"" <token> "\""
"""



import logging

from ..lib.job_control import Job
from .machine import Machine
from ..parser.parse import Parser


class ScriptRunner(Job):
    def __init__(self):
        self.program = None
        self.parser = Parser()
        self.vm = Machine()
        
    @classmethod
    def from_file(cls, file_name):
        new_instance = ScriptRunner()
        new_instance.load_file(file_name)
        return new_instance
    
    @classmethod
    def from_string(cls, script):
        new_instance = ScriptRunner()
        new_instance.load_string(script)
        return new_instance
        
    def load_file(self, file_name):
        self.program = self.parser.load(file_name)
        if self.program == None:
            logging.error(self.parser.get_errors())
        return self.program
    
    def load_string(self, input_string):
        self.program = self.parser.parse(input_string)
        if self.program == None:
            logging.error(self.parser.get_errors())
        return self.program
        
    def execute(self):
        if self.program != None:
            self.vm.run(self.program)

    def request_stop(self):
        self.vm.stop()


from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="bardolph",
    version="0.0.9",
    author="Al Fontes",
    author_email="bardolph@fontes.org",
    description="Simple script interpreter for LIFX light bulbs",
    url="https://www.bardolph.org",
    license = 'Apache License 2.0',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=[
        'bardolph.controller', 'bardolph.parser', 'bardolph.lib',
        'bardolph.fakes'],
    install_requires=['lifxlan'],
    python_requires='>=3.7',
    entry_points={
        'console_scripts': [
            'lsc=bardolph.controller:lsc.main',
            'lscap=bardolph.controller:snapshot.main',
            'lsrun=bardolph.controller:run.main',
            'lsparse=bardolph.parser:parse.main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha"
    ],
)



import collections
import logging
import threading


class Job:
    def execute(self): pass
    def request_stop(self): pass


class Promise:
    def __init__(self, job, callback):
        self.job = job
        self.callback = callback

    def execute(self):
        threading.Thread(
            target=self.execute_and_call).start()
        return self
              
    def execute_and_call(self):
        self.job.execute()
        return self.callback()

    def request_stop(self):
        self.job.request_stop()
    

class JobControl:
    def __init__(self, repeat=False):
        self.promise = None
        self.queue = collections.deque()
        self.repeat = repeat
        self.lock = threading.RLock()
    
    def set_repeat(self, repeat):
        self.repeat = repeat
    
    def add_job(self, job):
        return self.enqueue_job(job, self.queue.append)
    
    def insert_job(self, job):
        return self.enqueue_job(job, self.queue.appendleft)
                
    def enqueue_job(self, job, fn): 
        if not self.lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
        else:
            try:
                fn(job)
                if self.promise is None:
                    self.run_next_job()
            finally:
                self.lock.release()
    
    def has_jobs(self):
        return len(self.queue) > 0
                
    def run_next_job(self):
        if not self.lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
        else:
            try:
                if self.promise is None and len(self.queue) > 0:
                    next_job = self.queue.popleft()
                    if self.repeat:
                        self.queue.append(next_job) 
                    self.promise = Promise(next_job, self.on_execution_done)
                    self.promise.execute()
            finally:        
                self.lock.release()

    def on_execution_done(self):
        if not self.lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
        else:
            try:
                self.promise = None
                logging.debug("Script jobs finished.")
            finally:
                self.lock.release()        
            self.run_next_job()

    def request_stop(self):
        if not self.lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
        else:
            try:
                logging.debug("Stop requested.")
                self.repeat = False
                self.queue.clear()
                p = self.promise
                if p != None:
                    self.promise = None
                    p.request_stop()
            finally:
                self.lock.release()

    def request_finish(self):
        """ Clear out the queue but let the running job finish. """
        if not self.lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
        else:
            try:
                logging.debug("Finish requested.")
                self.repeat = False
                self.queue.clear()
            finally:
                self.lock.release()



import re

from .token_types import TokenTypes


class Lex:
    token_regex = re.compile(r'#.*$|".*?"|\S+') 
    number_regex = re.compile(r'^[0-9]*\.?[0-9]+$')

    def __init__(self, input_string):
        self.lines = iter(input_string.split('\n'))
        self.line_num = 0
        self.tokens = None
        self.next_line()
        
    def next_line(self):
        current_line = next(self.lines, None)
        if current_line is None:
            self.tokens = None
        else:
            self.line_num += 1
            self.tokens = self.token_regex.finditer(current_line)

    def unabbreviate(self, token):
        return {
            'h': 'hue', 's': 'saturation', 'b': 'brightness', 'k': 'kelvin'
        }.get(token, token)

    
    def next_token(self):
        token_type = None
        while token_type is None:
            match = next(self.tokens, None)
            while match is None:
                self.next_line()
                if self.tokens is None:
                    return (TokenTypes.EOF, '')
                else:
                    match = next(self.tokens, None)
            else:
                token = self.unabbreviate(
                    match.string[match.start():match.end()])
                if token[0] != '#':
                    if token[0] == '"':
                        token = token[1:-1]
                        token_type = TokenTypes.LITERAL
                    elif self.number_regex.search(token):
                        token_type = TokenTypes.NUMBER
                    else: 
                        token_type = TokenTypes.__members__.get(
                            token.upper(), TokenTypes.UNKNOWN)
        
        return (token_type, token)
